@echo off
setlocal EnableExtensions EnableDelayedExpansion
set path=%~dp0;%~dp0.\bin;%path%

REM +=================================================================+
REM | mkISO_RAMload_sort.cmd experimental v0.05                       |
REM | created by cdob                                                 |
REM |-----------------------------------------------------------------|
REM | NT based boot CD with sorted files                              |
REM | NT4,2000,XP,2003; 32bit,64bit                                   |
REM +=================================================================+


rem used external applications
rem mkisofs.exe : http://cdrecord.berlios.de  http://www.student.tugraz.at/thomas.plank/
rem dd.exe      : http://www.chrysocome.net/dd
rem gsar.exe    : http://home.online.no/~tjaberg

rem some defaults
set boot_sector_file=bootsect.bin
set IsoFile=%~dp0XP.iso
set SourceDir=
set mkisofs_option=

set boot_x86=I386
set boot_x64=AMD64
set boot_directory=

rem mkisofs sort file
set iso_level=3
set sort_lst=%~dp0sort.txt
if exist %sort_lst% del %sort_lst%
set path_txt=%~dp0pathlist.txt
if exist %path_txt% del %path_txt%

rem parse command line
:begin_parse
 set SourceDir=%~1
 if [%1]==[-o] (set IsoFile=%~f2&shift)
 if [%1]==[-volid] (set mkisofs_option=%mkisofs_option% %~1 "%~2"&shift)
 if [%1]==[-V] (set mkisofs_option=%mkisofs_option% %~1 "%~2"&shift)
 if [%1]==[-b] (set boot_sector_file=%~2&shift)
 if [%1]==[-iso-level] (set iso_level=%~2&shift)
 for %%a in (-r -R -J -joliet-long -v -boot-info-table -duplicates-once -force-uppercase) do (
   if [%1]==[%%a] set mkisofs_option=!mkisofs_option! %%a)
 if /I [%1]==[-H] (call :Help &goto :eof)
 if /I [%1]==[/H] (call :Help &goto :eof)
 if    [%1]==[-?] (call :Help &goto :eof)
 shift
 if [%1]==[] goto :exit_parse
goto begin_parse
:exit_parse

if not defined SourceDir (
 echo.&echo error: SourceDir not defined &call :Help &pause &goto :eof)
pushd "%SourceDir%"

if not exist %boot_sector_file% (
 echo.&echo error: boot sector file not found &call :Help &pause &goto :eof)

rem unify case - sort is case sensitive
rename %boot_sector_file% %boot_sector_file%
rename %boot_x86% %boot_x86%
if exist %boot_x86%\SYSTEM32 rename %boot_x86%\SYSTEM32 SYSTEM32
if exist %boot_x86%\SYSTEM32\$OEM$ rename %boot_x86%\SYSTEM32\$OEM$ $OEM$
if exist %boot_x86%\I386\ASMS rename %boot_x86%\I386\ASMS ASMS
if exist $OEM$ rename $OEM$ $OEM$
if exist $OEM$\TEXTMODE rename $OEM$\TEXTMODE TEXTMODE
if exist OEM rename OEM OEM
if exist AMD64 rename AMD64 AMD64
if exist AMD64 if %iso_level% LEQ 3 (
 set mkisofs_option=%mkisofs_option% -joliet-long -hide-joliet %boot_sector_file%)

rem 2003 SP2 long name
if exist %boot_x86%\SYSTEM.WEB.SECURITY.SINGLESIGNON.* if %iso_level% LEQ 3 (
  set mkisofs_option=%mkisofs_option% -joliet-long -hide-joliet %boot_sector_file%)
if exist %boot_x64%\SYSTEM.WEB.SECURITY.SINGLESIGNON.* if %iso_level% LEQ 3 (
  set mkisofs_option=%mkisofs_option% -joliet-long -hide-joliet %boot_sector_file%)

set mkisofs_option=%mkisofs_option% -iso-level %iso_level% -N -D -relaxed-filenames ^
 -b %boot_sector_file% -hide %boot_sector_file% -hide boot.catalog ^
 -no-emul-boot -boot-load-size 4 ^
 -sort "%sort_lst%"

call :mk_sort_list_dynamic

rem call :add_RAM_load_size

echo on
mkisofs.exe %mkisofs_option% -o "%IsoFile%" .
@echo off
popd

rem unresolved grub4dos questions
rem call :patch_sort_iso

rem call :list_RAM_load_size

call :patch_sort_iso_menu_lst

goto :eof =============================================================


:mk_sort_list_static ==================================================
rem basic static sort list - keep this
( 
echo ./boot.catalog 10000
echo ./%boot_sector_file% 9990
echo ./%boot_x64%/* 7500
echo ./%boot_x86%/* 7000
echo ./%boot_x86%/SETUPLDR.BIN 9988
echo ./%boot_x86%/BOOTFIX.BIN 9986
echo ./%boot_x86%/BOOTFONT.BIN 9984
echo ./%boot_x86%/NTDETECT.COM 9982
echo ./%boot_x86%/TXTSETUP.SIF 9980
echo ./WIN* 9978
echo ./OEM/* -5000
echo ./$OEM$/* 6000)>%sort_lst%

goto :eof =============================================================


:mk_sort_list_dynamic =================================================
echo create sort list

(echo ./boot.catalog 10000
 echo ./%boot_sector_file% 9990
 echo ./WIN* 9978
 echo ./win* 9978
 echo ./%boot_x64%/* 1100
 echo ./%boot_x64%/SYSTEM32 9000
 echo ./%boot_x64%/COMPDATA 20
 echo ./%boot_x64%/LANG 1000
 echo ./%boot_x64%/WINNTUPG 1000
 echo ./%boot_x86%/* 1100
 echo ./%boot_x86%/SYSTEM32 9000
 echo ./%boot_x86%/COMPDATA 20
 echo ./%boot_x86%/LANG 1000
 echo ./%boot_x86%/WIN9XMIG 1000
 echo ./%boot_x86%/WIN9XUPG 1000
 echo ./%boot_x86%/WINNTUPG 1000
 echo ./DOCS -1000
 echo ./DOTNETFX -1000)>>%sort_lst%
 
 
set weight_boot=9900
set weight_copy=8000

set boot_directory=
FOR %%a in (WIN* bootfont.bi?) do (call :sort_boot . %%a)

rem 32bit and 64bit OS
set boot_directory=%boot_x86%

rem 64bit OS
if exist %boot_x64%\txtsetup.sif set boot_directory=%boot_x64%

echo ./%boot_directory%/*.SY? 9010>>%sort_lst%

pushd "%SourceDir%\%boot_directory%"
if not exist txtsetup.sif (
  echo warning: %SourceDir%\%boot_directory%\txtsetup.sif not found &popd &goto :eof)

echo add copy sort files
echo add_copy_sort_files -1>>%sort_lst%
set sort_flag=
FOR /f "tokens=1,2* delims=.,=	 " %%a IN (txtsetup.sif) DO (
  rem echo %%a %%b %%c
  SET first_char=%%a
  IF "!first_char:~0,1!"=="[" SET sort_flag=
  IF /I "%%a"=="[SourceDisksFiles]" SET sort_flag=C
  IF /I "%%a"=="[SourceDisksFiles"  SET sort_flag=C
  
  set copy_file=
  IF defined sort_flag set copy_file=true
  
  rem not copied at textmode
  SET last_chars=%%c
  SET last_chars=!last_chars:~-2!
  if "!last_chars:~-2!"==",3" set copy_file=

  rem support compressed files; N.3 N.2 and N.1 names
  set name=%%a
  set ext=%%b
  set file=%%a.!ext:~0,-1!?
  rem N.2 and N.1 names
  if "!ext:~2,1!"=="" set file=%%a.%%b?
  rem N. names; compressed and expanded
  if not exist "!file!" if exist %%a set file=!name:~0,-1!?
  if not exist "!file!" if exist %%a.? set file=%%a.?
  
  if not exist "!file!" set copy_file=
  if defined copy_file call :sort_copy ./%boot_directory% !file!
)
rem more special cases
FOR %%a in (asms01.ca? ntld?) do (call :sort_copy ./%boot_directory% %%a)

rem flag file - end of RAM load
rem set RAMLoadEndFile=RAMBOOT.ISO
rem set RAMLoadEndMarker=#1#2#3#4#5#6#7#8#9#0%random%
rem echo %RAMLoadEndMarker%>%RAMLoadEndFile%
rem call :sort_copy ./%boot_directory% %RAMLoadEndFile%

set RAMLoadMenu=RAMBOOT.LST
set RAMLoadEndMarker=#1#2#3#4#5#6#7#8#9#0%random%
echo %RAMLoadEndMarker%> %RAMLoadMenu%
FOR /L %%a IN (1,1,50) DO echo #                                 >> %RAMLoadMenu%
call :sort_copy ./%boot_directory% %RAMLoadMenu%

rem -print-size -path-list
rem set RAMLoadTXT=RAMBOOT.TXT
rem call :sort_copy ./%boot_directory% %RAMLoadTXT%
rem set RAMLoadTXT=%boot_directory%\%RAMLoadTXT%

rem ###########################################################
echo add boot sort files
echo add_boot_sort_files -1>>%sort_lst%
FOR %%a in (spcmdcon.sy? txtsetup.si? biosinfo.in? ntkrnlmp.ex? ntoskrnl.ex? ^
  bootvid.dl? setupdd.sy? spddlang.sy? setupp.in? ^
  kdcom.dl? setupreg.hi? ctype.nl? locale.nl? l_intl.nl? sortkey.nl? ^
  sorttbls.nl? unicode.nl? c_437.nl? c_1252.nl? vgaoem.fo? ) do (call :sort_boot ./%boot_directory% %%a)

set sort_flag=off
FOR /f "tokens=1-3* delims=.,=	 " %%a IN (txtsetup.sif) DO (
  rem echo %%a %%b %%c
  SET first_char=%%a
  IF "!first_char:~0,1!"=="[" SET sort_flag=off
  IF /I "%%a"=="[files" SET sort_flag=A
  IF /I "%%b"=="Load]" SET sort_flag=B
  IF /I "%%a"=="[HiveInfs" SET sort_flag=B
  
  rem support compressed files
  set ext=%%b
  set file=%%a.!ext:~0,-1!?
  IF "!sort_flag!"=="A" if "%%c"=="4" if exist "!file!" call :sort_boot ./%boot_directory% !file!

  set ext=%%c
  set file=%%b.!ext:~0,-1!?
  IF "!sort_flag!"=="B" if exist "!file!" call :sort_boot ./%boot_directory% !file!
  
  IF /I "%%a"=="AnsiCodepage" call :sort_boot ./%boot_directory% !file!
  IF /I "%%a"=="OEMCodepage" call :sort_boot ./%boot_directory% !file!
  IF /I "%%a"=="UnicodeCasetable" call :sort_boot ./%boot_directory% !file!
  IF /I "%%a"=="OemHalFont" call :sort_boot ./%boot_directory% !file!
)

FOR %%a in (setupreg.hi? wmilib.sy? vga.sy? videoprt.sy* kbdclass.sy? classpnp.sy? ^
  scsiport.sy? storport.sy? drvmain.sd? ndis.sy? ipsec.sy? tcpip.sy? ^
  ipnat.sy? netbt.sy? rdbss.sy? mup.sy? mrxsmb.sy?) do (call :sort_boot ./%boot_directory% %%a)

if exist ASMS echo ./%boot_directory%/ASMS/* -7000>>%sort_lst%

popd

pushd "%SourceDir%\%boot_x86%"
FOR %%a in (bootfix.bi? setupldr.bi? win*.si? migrate.in? boot*.bi? ^
  ntdetect.co?) do (call :sort_boot ./%boot_directory% %%a)
popd

echo ./%boot_x86%/SYSTEM32/$OEM$/* %weight_boot%>>%sort_lst%>>%sort_lst%

(echo ./OEM/* 200
 echo ./$OEM$/* 100
 echo ./$OEM$/TEXTMODE/* 9950)>>%sort_lst%

goto :eof =============================================================


:sort_boot
  set /a weight_boot-=1
  for %%f in (%2) do (
    echo %1/%%f %weight_boot%>>%sort_lst%
    rem echo %1/%%f>>%path_txt%
  )
goto :eof

:sort_copy
  set /a weight_copy-=1
  for %%f in (%2) do (
    echo %1/%%f %weight_copy%>>%sort_lst%
    rem echo %1/%%f>>%path_txt%
  )
goto :eof


::=====================================================================
:Help
ECHO. &ECHO  %~nx0 [options] windows_source_directory
echo  -o : output ISO file name
echo  -b : boot sector file, relative path to windows source directory
echo  -V : set volume ID
echo.
goto :eof ============================================================= 


:patch_sort_iso =======================================================

rem warning: very experimental

rem dd.exe   : http://www.chrysocome.net/dd
rem gsar.exe : http://home.online.no/~tjaberg

rem ISO image with sorted files
rem filename: RAMBOOT.ISO
rem contents of RAMBOOT.ISO: #1#2#3#4#5#6#7#8#9#0

rem ISO9660: 9 File and Directory Descriptors
set OffsetFileIdentifier=
set DirectoryRecord=

set GsarString=-b -sRAMBOOT.ISO %IsoFile%
for /f "tokens=3 delims=: " %%a in ('gsar.exe %GsarString%^| find.exe /V "match found"') do set OffsetFileIdentifier=%%a

set /a DirectoryRecord = %OffsetFileIdentifier% - 34
set /a LocationOfExtent = %DirectoryRecord% + 3
set /a DataLength = %DirectoryRecord% + 11

echo OffsetFileIdentifier %OffsetFileIdentifier%
echo DirectoryRecord %DirectoryRecord%

if exist DirectoryRecord.bin del DirectoryRecord.bin
dd.exe if=%IsoFile% of=DirectoryRecord.bin skip=%DirectoryRecord% bs=1 count=48

rem how to read this binary file?
rem if exist LocationOfExtent.bin del LocationOfExtent.bin
rem dd.exe if=%IsoFile% of=LocationOfExtent.bin skip=%LocationOfExtent% bs=1 count=8

rem store some zero bytes to a file
if exist zero.bin del zero.bin
dd.exe if=%IsoFile% of=zero.bin bs=1 count=8

rem set LBA zero to DirectoryRecord.bin
dd.exe if=zero.bin of=DirectoryRecord.bin bs=1 seek=3 count=8

set RAMload=
set GsarString=-b -s%RAMLoadEndMarker% %IsoFile%
for /f "tokens=3 delims=: " %%a in ('gsar.exe %GsarString%^| find.exe /V "match found"') do set RamLoadSize=%%a

echo RAMload %RamLoadSize%

set FileSizeHex=0000000%RamLoadSize:~2%
echo %PE_ISO% file size: 0x%FileSizeHex%

set FileSizeHex=%FileSizeHex:~-8%
echo %PE_ISO% file size: 0x%FileSizeHex%

rem ISO9660: 9.1.4 Data Length stored according 7.3.3 Both-byte orders
set FileSizeISO=:x%FileSizeHex:~6,2%:x%FileSizeHex:~4,2%:x%FileSizeHex:~2,2%:x%FileSizeHex:~0,2%
set FileSizeISO=%FileSizeISO%:x%FileSizeHex:~0,2%:x%FileSizeHex:~2,2%:x%FileSizeHex:~4,2%:x%FileSizeHex:~6,2%
echo %FileSizeISO%

rem set FileSizeISO to DirectoryRecord.bin
gsar.exe -o -s:x00:x00:x00:x00:x00:x00:x00:x00 -r%FileSizeISO% zero.bin
dd.exe if=zero.bin of=DirectoryRecord.bin bs=1 seek=11 count=8

rem change the ISO file finally
dd.exe if=DirectoryRecord.bin of=%IsoFile% seek=%DirectoryRecord% bs=1 count=48

goto :eof


:list_RAM_load_size ===================================================

rem gsar.exe : http://home.online.no/~tjaberg

rem ISO image with sorted files
rem filename: RAMBOOT.ISO
rem contents of RAMBOOT.ISO: #1#2#3#4#5#6#7#8#9#0

set GsarString=-b -s%RAMLoadEndMarker% %IsoFile%
for /f "tokens=3 delims=: " %%a in ('gsar.exe %GsarString%^| find.exe /V "match found"') do set RamLoadSize=%%a
set /a RamLoadSectors = %RamLoadSize% / 2048

echo. &echo warning: very experimental
echo RAMload %RamLoadSize% bytes   %RamLoadSectors% sectors
echo.

goto :eof


:add_RAM_load_size ===================================================
rem simple approach -print-size -path-list, won't match final LBA

rem convert DOS newlines (CR/LF) to Unix format http://www.bastet.com/
dos2unix.exe  %path_txt%

set RamLoadSectors=

for /f "tokens=1" %%s in ('mkisofs.exe  -quiet %mkisofs_option% -print-size -path-list %path_txt%') do set RamLoadSectors=%%s

echo. &echo warning: very experimental
echo RAMload %RamLoadSectors% sectors
echo +%RamLoadSectors%>%RAMLoadTXT%
echo.

goto :eof


:patch_sort_iso_menu_lst ==============================================

rem warning: very experimental

rem dd.exe   : http://www.chrysocome.net/dd
rem gsar.exe : http://home.online.no/~tjaberg

rem ISO image with sorted files
rem filename: RAMBOOT.TXT
rem contents of RAMBOOT.TXT: #1#2#3#4#5#6#7#8#9#0

set RAMload=
set GsarString=-b -s%RAMLoadEndMarker% %IsoFile%
for /f "tokens=3 delims=: " %%a in ('gsar.exe %GsarString%^| find.exe /V "match found"') do set /a RamLoadSize=%%a
set /a RamLoadSectors = %RamLoadSize% / 2048

echo RAMload %RamLoadSize% RamLoadSectors %RamLoadSectors%

(echo #%RamLoadSectors%                                 #
 echo.
 echo timeout 1
 echo.
 echo title Install Windows - no map
 echo map --mem ^(0xFE^)+%RamLoadSectors% ^(0xFF^)
 echo map --hook
 echo chainloader ^(0xFF^)/I386/SETUPLDR.BIN
)>temp_RAMLOAD.LST

for %%A in (temp_RAMLOAD.LST) do set FileSize=%%~zA
echo FileSize %FileSize%

rem change the ISO file
dd.exe if=temp_RAMLOAD.LST of=%IsoFile% seek=%RamLoadSize% bs=1 count=%FileSize%
if exist temp_RAMLOAD.LST del temp_RAMLOAD.LST

goto :eof
